unit mt4_timezone_support_class;

{$mode objfpc}{$H+}

interface

uses
	Process,
	Classes,
	Windows,
	DateUtils,
	CMem,
	SysUtils,
	StrUtils,
	// Pascal-Timezones (source: http://wiki.freepascal.org/PascalTZ)
	upascaltz;
type
	TZSupportClass = class(TProcess)
		protected
		        tzInfo : TPascalTZ;
		        FBusyWaitCrtSect: TRTLCriticalSection;
			procedure Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring);
			procedure Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring; AArgs: array of const);
		public
			// start an instance of this class
			constructor Create(pTZInfoFileName : AnsiString); reintroduce;

			// stop the instance of this class. Call this before Free
			procedure Stop;

			// terminate the instance (don't call this, call Free as usual)
			destructor Destroy; override;

		        // Calculate the broker time by a time string and the broker timezone
		        function StrToBrokerTime(
		   		pCurrentBrokerDate	: LongInt;
		                pTimeStr		: AnsiString;
		                pBrokerTZ		: AnsiString
		        ):LONG;

		        // Returns TRUE when the pBrokerTime is between pTimeFrom and pTimeUntil
		        // Sets pTimeFrom and pTimeUntil to the respective Posix Timestamps
		        // according to pTimeFromStr and pTimeUntilStr
		        function isTimeBetweenRange(
		                pCurrentBrokerTime	: LongInt;
		                var	pTimeFrom	: LongInt;
				var	pTimeUntil	: LongInt;
		                pTimeFromStr		: AnsiString;
		                pTimeUntilStr		: AnsiString;
		                pBrokerTZ		: AnsiString
			):integer;
	end;

implementation

constructor TZSupportClass.Create(pTZInfoFileName : AnsiString);
begin
	inherited Create(nil);
	Msg(0, 'TZSupportClass', 'Creating');
	InitializeCriticalSection(FBusyWaitCrtSect);
        try
                Msg(0, 'TZSupportClass', 'Creating tzInfo object');
        	tzInfo := TPascalTZ.Create();
		tzInfo.ParseDataBaseFromFile(pTZInfoFileName);
                Msg(0, 'TZSupportClass', 'Created tzInfo object successfully');
        except
                Msg(0, 'TZSupportClass', 'ERROR while creating tzInfo object');
                tzInfo.Destroy();
                tzInfo:=NIL;
        end;
	Msg(0, 'TZSupportClass', 'Created');
end;

procedure TZSupportClass.Stop;
begin
  	Msg(-1, 'TZSupportClass', 'Destroying');
	DeleteCriticalSection(FBusyWaitCrtSect);
        Msg(-1, 'TZSupportClass', 'Destroyed');
end;

destructor TZSupportClass.Destroy;
begin
  self.Stop;
  if (tzInfo<>NIL) then
          tzInfo.Destroy;
  inherited Destroy;
end;


procedure TZSupportClass.Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring);
begin
	AMessage := '<' + IntToStr(ALevel) + '> ' + AMethod + ': ' + AMessage;
	{$ifdef win32}
	OutputDebugString(PChar(AMessage));
	{$else}
	writeln(AMessage);
	{$endif}
end;

procedure TZSupportClass.Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring; AArgs: array of const);
begin
	try
		Msg(ALevel, AMethod, Format(AMessage, AArgs));
	except
		Msg(ALevel, AMethod, 'Could not format message' );
	end;
end;


function TZSupportClass.StrToBrokerTime(
	pCurrentBrokerDate	: LongInt;
        pTimeStr		: AnsiString;
        pBrokerTZ		: AnsiString
):LONG;
var
        p	 : AnsiString;
        timeStr  : AnsiString;
	tzStr	 : AnsiString;
        tzSubFix : AnsiString;
        timeHour : AnsiString;
        timeMin  : AnsiString;
        i,j	 : Integer;
        rc	 : TDateTime;
        BrokerTime : TDateTime;
begin
	try
        	{
                Msg(0, 'StrToBrokerTime',
	        	'Invoking: par=("' + pTimeStr + '", "' + pBrokerTZ + '")'
                );
                }
		timeStr:=trim(pTimeStr);
	        if (length(timeStr)=0) then
	        	exit(-1);
                BrokerTime := DateOf(UnixToDateTime(pCurrentBrokerDate));
                // TODO: Check whether we've already calculated this date
                // to avoid costly searches in the TZ Database.


	        // Split the time zone information from the time string
	        tzStr := '';
	        timeHour := '';
	        timeMin := '';
	        // We expect a time string "hh24:mi timezone"
	        i:=1;
	        j:=1;
	        p:=timeStr;
	        while (p[i]>='0') and (p[i]<='9') do begin
	                i:=i+1;
	        end;
	        timeHour:=midStr(p, 1, i-j);
	        i:=i+1;
	        j:=i;
	        while (p[i]>='0') and (p[i]<='9') do begin
	                i:=i+1;
	        end;
	        timeMin:=midStr(p, j, i-j);
	        if (i<length(p)) then begin
	                tzStr:=trim(midStr(p, i, length(p) - i + 1));
	        end;
	        p:=timeHour+':'+timeMin;
                {
                Msg(0, 'StrToBrokerTime',
	        	'Time: ' + p + ' ' + tzStr
                );
                }
                // RC now contains a timezone free representation.
	        rc:=StrToDateTime(p);
	        //
	        // Add the current date from the broker
	        rc:= rc + BrokerTime;
                {
                Msg(0, 'StrToBrokerTime',
	        	'Time OLD: ' + DateTimeToStr(rc) + ' ' + tzStr
                );
                }
	        // and recalculate rc to the target time zone
		if (length(tzStr)>0) and (tzInfo<>NIL) then begin
	                rc:=tzInfo.TimeZoneToTimeZone(rc, tzStr, pBrokerTZ, tzSubFix);
	        end;
                {
                Msg(0, 'StrToBrokerTime',
	        	'Time NEW: ' + DateTimeToStr(rc) + ' ' + pBrokerTZ
                );
                }
	        // Return a Posix-Timestamp compatible with MT4
	        exit( LONG(Round((rc - 25569) * 86400)) );
        except
                Msg(0, 'StrToBrokerTime',
	        	'Error calculating broker time par=("' + pTimeStr + '", "' + pBrokerTZ + '")'
                );
          	exit(-1);
        end;
end;

function TZSupportClass.isTimeBetweenRange(
                pCurrentBrokerTime	: LongInt;
                var	pTimeFrom	: LongInt;
		var	pTimeUntil	: LongInt;
                pTimeFromStr		: AnsiString;
                pTimeUntilStr		: AnsiString;
                pBrokerTZ		: AnsiString
	):integer;
var
        rc	:integer;
        t	:LongInt;
begin
        { Msg(0, 'isTimeBetweenRange',
        	'Invoking CalculateBrokerTime: par=("' + pTimeFromStr + '", "' + pBrokerTZ + '")'
        ); }
        t:=StrToBrokerTime(pCurrentBrokerTime, pTimeFromStr, pBrokerTZ);
	{ Msg(0, 'isTimeBetweenRange',
        	'Setting pTimeFrom to  %d ',[t]
        ); }
        pTimeFrom  := t;
        { Msg(0, 'isTimeBetweenRange',
        	'Invoking CalculateBrokerTime: par=("' + pTimeUntilStr + '", "' + pBrokerTZ + '")'
        ); }
        t:=StrToBrokerTime(pCurrentBrokerTime, pTimeUntilStr, pBrokerTZ);
        { Msg(0, 'isTimeBetweenRange',
        	'Setting pTimeUntil to %d ',[t]
        ); }
        pTimeUntil :=t;
        // Add 24 hours to the UNTIL-Time to get a valid range
        if (pTimeFrom>pTimeUntil) then begin
                pTimeUntil:=pTimeUntil + 3600*24;
        end;
	if ( (pCurrentBrokerTime>=pTimeFrom) and (pCurrentBrokerTime<pTimeUntil) ) then begin
                rc:=1;
        end else begin
        	rc:=0;
        end;
        exit(rc);
end;



end.

